package czlib

import (
	"fmt"
	"io"
)

type reader struct {
	r      io.Reader
	in     []byte
	strm   *zstream
	err    error
	skipIn bool
}

// NewReader with default options
func NewReader(r io.Reader) (io.ReadCloser, error) {
	return NewReaderCustom(r, MAX_WBITS)
}

// NewReaderCustom for specifying window bits flag
func NewReaderCustom(r io.Reader, windowBits int) (io.ReadCloser, error) {
	return NewReaderCustomBuffer(r, windowBits, DEFAULT_COMPRESSED_BUFFER_SIZE)
}

// NewReaderCustomBuffer for specifying window bits flag and buffer size
func NewReaderCustomBuffer(r io.Reader, windowBits, bufferSize int) (io.ReadCloser, error) {
	z := &reader{r: r, in: make([]byte, bufferSize), strm: zstreamAlloc()}
	if err := z.strm.inflateInit(windowBits); err != nil {
		return nil, err
	}
	return z, nil
}

// Read bytes for decompression
func (z *reader) Read(p []byte) (int, error) {
	if z.err != nil {
		return 0, z.err
	} else if len(p) == 0 {
		return 0, nil
	}

	z.strm.setOutBuf(p, len(p))

	for {
		if !z.skipIn && z.strm.availIn() == 0 {
			var n int
			n, z.err = z.r.Read(z.in)

			if n > 0 && z.err == io.EOF {
				z.err = nil
			}

			if (z.err != nil && z.err != io.EOF) || (n == 0 && z.err == io.EOF) {
				z.strm.inflateEnd()
				return 0, z.err
			}

			z.strm.setInBuf(z.in, n)
		} else {
			z.skipIn = false
		}

		ret, err := z.strm.inflate(Z_NO_FLUSH)
		if err != nil {
			z.err = err
			z.strm.inflateEnd()
			return 0, z.err
		}

		have := len(p) - z.strm.availOut()
		if have > 0 {
			z.skipIn = ret == Z_OK && z.strm.availOut() == 0
			return have, z.err
		}
	}
}

// Close reader
func (z *reader) Close() error {
	if z.strm == nil {
		panic(fmt.Errorf("czlib: stream closed"))
	}
	defer func() {
		z.strm.inflateEnd()
		zstreamFree(z.strm)
		z.strm = nil
	}()

	if z.err != io.EOF {
		return z.err
	}

	z.err = io.EOF
	return nil
}
