package czlib

import (
	"fmt"
	"io"
)

// DEFAULT_COMPRESSED_BUFFER_SIZE while compressing/decompressing
const DEFAULT_COMPRESSED_BUFFER_SIZE = 32 * 1024

// Writer is intended for compression logic
type Writer struct {
	w    io.Writer
	out  []byte
	strm *zstream
	err  error
}

// NewWriter with default options
func NewWriter(w io.Writer) *Writer {
	z, _ := NewWriterCustom(w, Z_DEFAULT_COMPRESSION, Z_DEFLATED, MAX_WBITS, MAX_MEM_LEVEL, Z_DEFAULT_STRATEGY)
	return z
}

// NewWriterCustom for specifying custom zlib flags
func NewWriterCustom(w io.Writer, level, method, windowBits, memLevel, strategy int) (*Writer, error) {
	return NewWriterCustomBuffer(w, level, method, windowBits, memLevel, strategy, DEFAULT_COMPRESSED_BUFFER_SIZE)
}

// NewWriterCustomBuffer for specifying custom zlib flags and buffer size
func NewWriterCustomBuffer(w io.Writer, level, method, windowBits, memLevel, strategy, bufferSize int) (*Writer, error) {
	z := &Writer{w: w, out: make([]byte, bufferSize), strm: zstreamAlloc()}
	if err := z.strm.deflateInit(level, method, windowBits, memLevel, strategy); err != nil {
		return nil, err
	}
	return z, nil
}

func (z *Writer) write(p []byte, flush int) int {
	if len(p) == 0 {
		z.strm.setInBuf(nil, 0)
	} else {
		z.strm.setInBuf(p, len(p))
	}

	for {
		z.strm.setOutBuf(z.out, len(z.out))
		z.strm.deflate(flush)

		from := 0
		have := len(z.out) - int(z.strm.availOut())
		for have > 0 {
			var n int
			n, z.err = z.w.Write(z.out[from:have])
			if z.err != nil {
				z.strm.deflateEnd()
				return 0
			}
			from += n
			have -= n
		}

		if z.strm.availOut() != 0 {
			break
		}
	}

	if z.strm.availIn() != 0 {
		panic(fmt.Errorf("czlib: unexpected error (2)"))
	}

	return len(p)
}

// Write bytes for compression
func (z *Writer) Write(p []byte) (n int, err error) {
	if z.err != nil {
		return 0, z.err
	}
	n = z.write(p, Z_NO_FLUSH)
	return n, z.err
}

// Flush contents for compression
func (z *Writer) Flush() error {
	if z.err != nil {
		return z.err
	}
	z.write(nil, Z_SYNC_FLUSH)
	return z.err
}

// Close writer
func (z *Writer) Close() error {
	if z.strm == nil {
		panic(fmt.Errorf("czlib: stream closed"))
	}
	defer func() {
		z.strm.deflateEnd()
		zstreamFree(z.strm)
		z.strm = nil
		z.err = io.EOF
	}()

	if z.err != nil {
		return z.err
	}
	z.write(nil, Z_FINISH)
	if z.err != nil {
		return z.err
	}

	return nil
}
